#include<iostream>
#include<fstream>
#include<cstdio>
#include<cstdlib>
#include<cstring>
#include<cassert>
#include<string>
#include<stdint.h>
#include<limits>
#include<map>
#include<vector>
#include<sstream>
#include<time.h>
#include <algorithm>
#include <iterator>
#include "misc.h"
#include "SubmatClass.h"
#include "nullModel.h"
#include "dpcellClass.h"
#include "threeStateDPAClass.h"
#include "cmdLineParser.h"
#include "aaAlphabet.h"

using namespace std;

struct SSTMapping{
    string sequence = "";
    string sseStr = "";
    string header;
};

vector<string> readFastaFile(char fname[]) {
  vector<string> seqs;

  ifstream infile(fname,ios::in);
  assert(infile);
  string astr;
  char buff[10000];
  size_t cntr = 0 ;
  while (!infile.eof()) {
    infile.getline(buff,10000);
    if (infile.eof()) {
        seqs.push_back(astr);
        break;
    }
    if (buff[0] == '>')  {
      if (cntr > 0) {
        seqs.push_back(astr);
      }
      cntr++;
      astr = "";
    }
    else {
      astr.append(buff);
    }
  }
  infile.close();
 return seqs;
}

vector<string> getAASeqs(vector<string> fn) { //expects files in fasta format
    vector<string> s1 = readFastaFile((char *)fn[0].c_str());
    vector<string> s2 = readFastaFile((char *)fn[1].c_str());
    assert(s1.size() == 1);
    assert(s2.size() == 1);
    int n = s1.size()+s2.size();
    if (n != 2) {
        cerr << "Expecting 2 amino acid sequences. Found " << n << "instead. Fix!\n";
        exit(1);
    }
    vector<string> s;
    s.push_back(s1[0]);
    s.push_back(s2[0]);
    return s;
}

vector<string> removeNonStandardAA(string s, string ssestr) {
    assert(s.length() == ssestr.length());
    string sanitized_s, sanitized_sse;
    for (size_t i = 0; i < s.length(); i++) {
        switch (s[i]) {
            case 'A':
            case 'C':
            case 'D':
            case 'E':
            case 'F':
            case 'G':
            case 'H':
            case 'I':
            case 'K':
            case 'L':
            case 'M':
            case 'N':
            case 'P':
            case 'Q':
            case 'R':
            case 'S':
            case 'T':
            case 'V':
            case 'W':
            case 'Y': sanitized_s.push_back(s[i]);
                      sanitized_sse.push_back(ssestr[i]);
                      break;
            default: //cout << "ignoring " << s[i] << " at pos = " << i+1 << endl;
                     break;
        }
    }
    return {sanitized_s, sanitized_sse};
}

double getNullEncodingLength(string s, NullModel& nullMdl) {
    double cdlen = 0;
    size_t order = nullMdl.getOrder();
    size_t alphabetSize = nullMdl.getAlphabetSize();

    //Encode symbols without full context uniformly
    //over the alphabet
    double p;
    for (int i = 0; i < order; i++) {
        p = (double)1/alphabetSize;
        cdlen += -log(p); 
        //cout << p << " " << nits2bits(-log(p)) << " " << nits2bits(cdlen) << endl;
    }
    for (int i = order; i < s.length(); i++) {
        p = nullMdl.pr(s.substr(i-order,order+1));
        cdlen += -log(p);
        //cout << s.substr(i-order,order+1) << " --> " 
        //     << p << " " << nits2bits(-log(p)) << " " << nits2bits(cdlen) << endl;
    }
    return cdlen;
}

SSTMapping parseSST(string sstFilename){
    SSTMapping sst;
    sst.sequence = "";
    sst.sseStr = "";

    string pdb = getFilenameFromPath(sstFilename);

    std::fstream sst_file(sstFilename.c_str(), std::fstream::in);

    if(sst_file.fail()) {
        cerr << "Failed to open file " << sstFilename << endl;
        exit(-1);
    }

    string seq = "", ssestr = "";
    char chain = ' ', chainSS;

    while(!sst_file.eof()){

        string current_line, current_word;
        getline(sst_file, current_line);

        if(current_line[0] == '+'){
            stringstream line_ss;
            line_ss.str(current_line);
            line_ss.ignore(256, '+');
            line_ss >> chainSS;

            if(!(isblank(chain) || chain == chainSS)){
                assert(seq.length() == ssestr.length());
                assert(!isblank(chain));
                assert(seq.length() > 0);

                vector<string> sanitizd = removeNonStandardAA(seq, ssestr);
                if(sanitizd[0].length() > 0){
                    sst.sequence += sanitizd[0];
                    sst.sseStr += sanitizd[1];
                    sst.header = pdb;
                }

                seq = "";
                ssestr = "";
                chain = ' ';
            }

            chain = (chainSS != '(') ? chainSS : ' ';
            line_ss.ignore(10, ')');
            line_ss >> current_word;
            seq += current_word;

        } else if(current_line[0] == '-'){
            stringstream line_ss;
            line_ss.str(current_line);
            line_ss.ignore(256, '-');
            line_ss >> current_word;
            ssestr += current_word;
        }

        if (sst_file.peek() == '#') {
            assert(seq.length() == ssestr.length());
            assert(seq.length() > 0);

            vector<string> sanitizd = removeNonStandardAA(seq, ssestr);
            if(sanitizd[0].length() > 0){
                sst.sequence += sanitizd[0];
                sst.sseStr += sanitizd[1];
                sst.header = pdb;
            }
            
            break;
        }
    }

    return sst;
}    

double computeINullST( string &S, string &T,
 NullModel &nullMdl, double &lensaving, double &adaptivesaving) {

  size_t nS = S.size();
  size_t nT = T.size();

  vector<double> normwtcdlens = getWTCodelensNormalized_0_to_N_nits(nS+nT);
  int diff = nS-nT;
  if (diff < 0) diff *= -1;
  //cout << nS << " " << nT << " " << diff << endl;
  //cout << "New computation: " << nits2bits(wtCodeLen_nits(nS+nT)+ normwtcdens[diff]) << endl;


  //regular
  double nullModelMsgLen_S = getNullEncodingLength(S, nullMdl);
  double nullModelMsgLen_T = getNullEncodingLength(T, nullMdl);

  vector<string> seqs;
  seqs.push_back(S);
  seqs.push_back(T);
  /*
  //adaptive
  double adaptiveMsgLen = getAdaptiveEncoding(nullMdl,seqs);
  adaptivesaving = nullModelMsgLen_S+nullModelMsgLen_T-adaptiveMsgLen;
  */

  //mmlests
  //adaptNulltoMMLEstimates(nullMdl,seqs);
  double mmlestMsgLen = getNullEncodingLength(S, nullMdl) + getNullEncodingLength(T, nullMdl);
  


  //encode lengths
  double combinedLengths_msglen = wtCodeLen_nits(nS+nT)+ normwtcdlens[diff];
  lensaving =  (wtCodeLen_nits(nS)+wtCodeLen_nits(nT))-combinedLengths_msglen;

  return combinedLengths_msglen+nullModelMsgLen_S+nullModelMsgLen_T;
}

SSTMapping cleanUpSeqs(string& align, SSTMapping& sst){

    SSTMapping updated;
    updated.header = sst.header;
    updated.sequence = "";
    updated.sseStr = "";


    string seq = "";
    for(int i = 0; i < align.length(); i++){
        if(align[i] == '-') continue;
        seq += align[i];
    }

    if(sst.sequence.length() == seq.length() &&
       sst.sequence == seq) return sst;

    assert(sst.sequence.length() >= seq.length());   

    int j = 0;
    for(int i = 0; i < seq.length(); i++){
       
        while(seq[i] != sst.sequence[j]){
            j++;
        }
        updated.sequence += sst.sequence[j];
        updated.sseStr += sst.sseStr[j];
    }

    assert(updated.sequence == seq);

    return updated;
}

int main( int argc , char *argv[] ) {
    
    Parser inp = Parser(argc, argv);
    
    vector<string> param_seqfn;
    string param_nullProbFile = "data/sse_nullEstimates_SCOP259092_GapRegion.txt";
    string alphfile = "data/sseAlphabet.txt";
   
    AAAlphabet ssealph(alphfile);

    /* get SubmatClass_t obj */
    SubmatClass_t submat(ssealph); // instantiate MMLSUM series condl and joint matrices 
 
    /* get nullModelClass_t obj */
    NullModel nullMdl(param_nullProbFile);
    assert(nullMdl.isCompatible(submat.getAlphabet()));

    vector<string> ssealign;

    if(inp.fflag){
        ssealign = readFastaFile((char*)inp.sstalignf.c_str());

    } else {   
        SSTMapping sstp1 = parseSST(inp.sstf1);
        SSTMapping sstp2 = parseSST(inp.sstf2);

        vector<string> align = readFastaFile((char*)inp.stralignf.c_str());
        assert(align.size() == 2);
        assert(align[0].length() == align[1].length());

        SSTMapping sst1 = cleanUpSeqs(align[0], sstp1);
        SSTMapping sst2 = cleanUpSeqs(align[1], sstp2);

        ssealign = vector<string>(2, string(align[0].length(), '-'));

        int fi = 0, se = 0;
        for(int i = 0; i < align[0].length(); i++){
            if(align[0][i] != '-'){
                assert(align[0][i] == sst1.sequence[fi]);
                ssealign[0][i] = sst1.sseStr[fi];
                fi++;
            }

            if(align[1][i] != '-'){
                assert(align[1][i] == sst2.sequence[se]);
                ssealign[1][i] = sst2.sseStr[se];
                se++;
            }
        }
    }
   
        
    assert (ssealign.size() == 2);
    assert (ssealign[0].size() == ssealign[1].size());

    for(int i = 0; i < ssealign[0].size(); i++){
        if(ssealign[0][i] == '-' && ssealign[1][i] == '-'){
            ssealign[0].erase(ssealign[0].begin() + i); 
            ssealign[1].erase(ssealign[1].begin() + i); 
        }
    }
    
    assert (ssealign[0].size() == ssealign[1].size());

    vector<string> sanitized_algn = {ssealph(ssealign[0], true), ssealph(ssealign[1], true)};
    string algnfsastr = algn2Fsastr(sanitized_algn);
    string S = ssealph(sanitized_algn[0], false);
    string T = ssealph(sanitized_algn[1], false);

    cout << "----------------------------+=o0o=+-------------------------------" << endl;
    printSeq2Term(sanitized_algn[0],"Alignment String of S:");
    cout << endl;
    printSeq2Term(sanitized_algn[1],"Alignment String of T:");

    /* prepare ivalue stats*/
    threeStateDPAClass_t ivalueobj(S,T,algnfsastr,nullMdl,submat);

    double algnModelMsgLen = ivalueobj.getOptAlignMsgLen();

    double lensaving,adaptivesaving;
    double nullModelMesgLen = computeINullST(S,T,nullMdl,lensaving,adaptivesaving);
    double compression = nullModelMesgLen - algnModelMsgLen;
    cout << "\n----------------------------+=o0o=+-------------------------------\n";
    cout <<  getTerminalColorString("Inferred structure time =   ",45) << ivalueobj.getInferredSubmatN()  << endl;
    cout << "----------------------------+=o0o=+-------------------------------\n";
      
    return 0;
}

