#include "nullModel.h"

NullModel::NullModel() { }

/* Assumes fname file holds the null probabilities in a 
* matrix space where
* First column holds a string containing the context (i)
* First row holds the letter whose null model probability (j)
* is shown in the relevant (i,j) cell
*
* Read fname file under this assumption*/
NullModel::NullModel(string fname) {

    ifstream infile(fname.c_str(),ios::in);
    assert(infile);
    char buff[10000];
    vector<string> inpalph;
    vector<double> tempprobs;

    while (!infile.eof()) {

        infile.getline(buff,10000);
        stringstream ss;

        if(buff[0] == '#') {
            ss << buff;

            char colResidue;
            ss.ignore(20, ' ');
            while(ss >> colResidue)
                inpalph.push_back(string(1, colResidue));

            alphabet = AAAlphabet(inpalph);   
            tempprobs = vector<double>(alphabet.size());
            continue;
        }

        ss << buff;
        string c;
        double p;

        ss >> c; //context: first column

        int k = 0;
        if(isdigit(c[0])){
            int indx = alphabet[inpalph[k++]];
            assert(indx >= 0 && indx < tempprobs.size());
            tempprobs[indx] = stod(c);
            c = "";
        }

        while(ss >> p){ // rest of the columns
            int indx = alphabet[inpalph[k++]];
            assert(indx >= 0 && indx < tempprobs.size());
            tempprobs[indx] = p;
        }

        normalize(tempprobs);
        if(c.length() == 0){
            ctx2pr.push_back(tempprobs);
            order = 0;
        } else {
            int veclen = pow(ALPHABET_MAX_SIZE, c.length());
            if(ctx2pr.size() < veclen){
                ctx2pr = vector<vector<double>>(veclen);
            } 
            int ctxindx = alphabet[c];
            assert(ctxindx >= 0 && ctxindx < ctx2pr.size());
            ctx2pr[ctxindx] = tempprobs;
            order = c.length();
        }

        if (infile.eof() == true) break;

    }
    infile.close();

    /* ensure that the file contains all probabilities needed 
    * for the given alphabet */
    //sanityCheckProbs();
}

void NullModel::sanityCheckProbs(){
    double prob, sum = 0;
    if(order == 0){
        for(char c : alphabet.getAlphabet()){
            cout << setw(6) << string(1, c) << " ";
        } cout << "  Sum" << endl;

        for(char c : alphabet.getAlphabet()){
            prob = pr(string(1,c));
            sum += prob;
            cout << fixed << setprecision(4) 
                 << prob << " ";
        } cout << sum << endl;

    } else{
        string p(order, ' ');
        vector<string> perm;
        string alph = alphabet.getAlphabet();
        allPermutations(p, 0, alph, perm);

        cout << string(order, ' ');
        for(char c : alphabet.getAlphabet()){
            cout << setw(6) << string(1, c) << " ";
        } cout << "  Sum" << endl;
        for(string cntxt : perm){
            sum = 0;
            int row = alphabet[cntxt];
            assert(row >= 0 && row < ctx2pr.size());
            cout << cntxt<< " ";
            for(char c : alphabet.getAlphabet()){
                prob = pr(cntxt + string(1,c));
                sum += prob;
                cout << fixed << setprecision(4) 
                    << prob << " ";
            } cout << sum << endl;
        }
    }
}

void NullModel::allPermutations(string& permutations, int indx, string& alph, vector<string>& perms){
   for (int i = 0; i < alph.length(); i++) {
        permutations[indx] = alph[i];
        if (indx == order - 1){
           perms.push_back(permutations);
        } else
            allPermutations(permutations, indx+1, alph, perms);
   }
}

void NullModel::normalize(vector<double>& vec){

    double sum = accumulate(vec.begin(), vec.end(), 0.0);
    for(int i = 0; i < vec.size(); i++){
        vec[i] /= sum;
    } 

    double test = 0;
    for(int i = 0; i < vec.size(); i++){
        test += vec[i];
    }
    assert(test > 0.999999 && test <= 1.000000001);
}

double NullModel::pr(string ctx) {
    assert(ctx.length() - 1 == order); 
    int col = alphabet[ctx.back()];
    ctx.pop_back();
    int row = alphabet[ctx];

    assert(row >= 0 && row < ctx2pr.size());
    assert(col >= 0 && col < ctx2pr[row].size());
    return ctx2pr[row][col];
}

size_t NullModel::getOrder() {
    return order;
}

string NullModel::getAlphabet() {
    return alphabet.getAlphabet();
}

int NullModel::getAlphabetSize() {
    return alphabet.size();
}

bool NullModel::isCompatible(string otherAlphabet) {
    
    for (int i = 0; i < otherAlphabet.length(); i++) {
        size_t found = alphabet.getAlphabet().find(otherAlphabet[i]);
        if (found == string::npos) return false;
    }
    return true;
}
