#include "threeStateDPAClass.h"
#define HTML_HEADER "<html><head></head><body>"
#define HTML_FOOTER "</body></html>"
#define BEGIN_TABLE "<table border=\"1\">"
#define END_TABLE "</table>"
#define BEGIN_ROW "<tr>"
#define END_ROW "</tr>"
#define BEGIN_CELL "<td>"
#define BEGIN_COLOURED_CELL(c) ("<td style=\"background-color:" c "\">")
#define END_CELL "</td>"


using namespace std;

//this interface to compute ivalue given an fsastring
threeStateDPAClass_t::threeStateDPAClass_t(
 string &S, 
 string &T,
 string &fsastr,
 NullModel &nullMdl,
 SubmatClass_t &submat
) {
    this->S = S;
    this->T = T;
    this->nullMdl = nullMdl;
    this->submat = submat;
    
    runstat = false;

    double nS = S.length();
    double nT = T.length();
    assert(nS>0);
    assert(nT>0);

    estimateSubmatParam(fsastr);
    initParams();
    paramStatementMsglen = 0;
   
    if(matchOnly)
        algnModelMsgLen = fsa2ISTgAMatchOnly(fsastr);
    else 
        algnModelMsgLen = fsa2IA(fsastr) + fsa2ISTgA(fsastr);

    runstat = true;
}

void threeStateDPAClass_t::estimateSubmatParam(string fsastr) {
    if (fsastr.length() == 0) return;

    //cout << "Estimating submat param for fsastr:\n" << fsastr << endl;
    vector<string> Amatches; // note, vector of 2-tuples representing match cols.
    vector<string> A;        // full alignment cols.
    int iS = 0, jT = 0;
    for (size_t l = 0; l < fsastr.length(); l++) {
        //cout << fsastr[l] << " ";
        if (fsastr[l] == 'd') {
            string mpair = "";
            mpair += S[iS];
            mpair += '-';
            assert(mpair.length() == 2);
            A.push_back(mpair);
            iS++;
        }
        else if (fsastr[l] == 'i') {
            string mpair = "";
            mpair += '-';
            mpair += T[jT];
            assert(mpair.length() == 2);
            A.push_back(mpair);
            jT++;
        }
        else if (fsastr[l] == 'm') {
            string mpair = "";
            mpair += S[iS];
            mpair += T[jT];
            assert(mpair.length() == 2);
            A.push_back(mpair);
            Amatches.push_back(mpair); // separately, only matches
            iS++;
            jT++;
        }
        else {
            cout << fsastr[l] << " " << "Something wierd!" << endl;
            assert(1);
        }
    }

    double min = numeric_limits <double>::infinity();
    int min_k = MARKOV_TIME_LLIMIT;
 
    int lo = MARKOV_TIME_LLIMIT;
    int hi = MARKOV_TIME_ULIMIT; 

    min_k = MARKOV_TIME_LLIMIT;
    min = numeric_limits <double>::infinity();
    if(matchOnly){
        for(int K = MARKOV_TIME_LLIMIT; K <= MARKOV_TIME_ULIMIT; K++){
            double msglen = evalNegLogMatchesOnly_fullAlignment(fsastr,A,K);
            if (msglen < min) {
                min = msglen;
                min_k = K;
            }
        }
    } else {
        for(int K = MARKOV_TIME_LLIMIT; K <= MARKOV_TIME_ULIMIT; K++){
       
            double msglen = evalNegLogMatches_fullAlignment(fsastr,A,K);
            if (msglen < min) {
                min = msglen;
                min_k = K;
            }

        }
    }

    inferredSubmat_N = min_k;
}

void threeStateDPAClass_t::initParams() {
   assert(fsaPr.size() == 0);
 
   double prMM = submat.getprMM(inferredSubmat_N);
   double prMI = 0.5*(1.0-prMM);
   double prMD = 0.5*(1.0-prMM);
   fsaPr.insert(pair<string,double>("MM",prMM));
   fsaPr.insert(pair<string,double>("MI",prMI));
   fsaPr.insert(pair<string,double>("MD",prMD));

   //double prII = Imode[0];
   double prII = submat.getprII(inferredSubmat_N);
   double prDD = prII;
   //double prIM = Imode[1];
   double prIM = submat.getprIM(inferredSubmat_N);
   double prDM = prIM;
   double prID = 1.0-prII-prIM;
   double prDI = prID;

   fsaPr.insert(pair<string,double>("IM",prIM));
   fsaPr.insert(pair<string,double>("II",prII));
   fsaPr.insert(pair<string,double>("ID",prID));

   fsaPr.insert(pair<string,double>("DM",prDM));
   fsaPr.insert(pair<string,double>("DI",prDI));
   fsaPr.insert(pair<string,double>("DD",prDD));
}

void threeStateDPAClass_t::estimateParams(string fsastr) {
  paramStatementMsglen = log(MARKOV_TIME_ULIMIT - MARKOV_TIME_LLIMIT + 1);  

  if (fsastr.length() > 0) {
      estimateSubmatParam(fsastr);
      estimate3StateMachineParams_dirichlet(inferredSubmat_N);
  }
}

double threeStateDPAClass_t::evalNegLogMatches(vector<string> A, int n) {
    double val = 0;
    for (int l = 0; l < A.size(); l++) {
        double term = 0;
        for (int k = 0; k < 20; k++) {
            int i = submat.getIndx(A[l][0]);
            int j = submat.getIndx(A[l][1]);
            double     lmda = submat.getD(k,k);
            double    S_ik  = submat.getS(i,k);
            double invS_kj  = submat.getSinv(k,j);
            term += pow(lmda,n)*S_ik*invS_kj; 
        }
        val += -log(term);
    }
    return val;
}

double threeStateDPAClass_t::evalNegLogMatches_fullAlignment(string fsastr, vector<string> A, int n) {
    
    assert(fsastr.length() == A.size());
    assert(n >= MARKOV_TIME_LLIMIT);
    assert(n <= MARKOV_TIME_ULIMIT);

    double prMM = submat.getprMM(n);
    double prMI = 0.5*(1.0-prMM);
    double prMD = 0.5*(1.0-prMM);

    //double prII = Imode[0];
    double prII = submat.getprII(n);
    double prDD = prII;
    //double prIM = Imode[1];
    double prIM = submat.getprIM(n);
    double prDM = prIM;
    double prID = 1.0-prII-prIM;
    double prDI = prID;

    double msglen = log(MARKOV_TIME_ULIMIT - MARKOV_TIME_LLIMIT + 1);
    msglen += wtCodeLen_nits(fsastr.length());
    //first aligned pair
    char prev_state = '*';
    for (int l = 0; l < A.size(); l++) {
        char curr_state = fsastr[l];  
        if (curr_state == 'm') {
            assert(A[l][0]!='-'&&A[l][1]!='-');
            char S_i = A[l][0];
            char T_j = A[l][1];
            double matchPr = submat.jointpr(S_i,T_j,n);
            msglen += -log(matchPr);
        }
        else if (curr_state == 'd') {
            assert(A[l][0]!='-'&&A[l][1]=='-');
            char S_i = A[l][0];
            double nullPr_S_i = nullMdl.pr(string(1,S_i));
            msglen += -log(nullPr_S_i);
        }
        else if (curr_state == 'i') {
            assert(A[l][0]=='-'&&A[l][1]!='-');
            char T_j = A[l][1];
            double nullPr_T_j = nullMdl.pr(string(1,T_j));
            msglen += -log(nullPr_T_j);
        }
        else {
            assert(!"Non m/i/d state found in fsastr");
        }
       

        if (prev_state == '*') msglen += -log(1.0/3.0);
        else {
                if (prev_state == 'm' && curr_state == 'm') msglen += -log(prMM);
            else if (prev_state == 'm' && curr_state == 'd') msglen += -log(prMD);
            else if (prev_state == 'm' && curr_state == 'i') msglen += -log(prMI);
            else if (prev_state == 'd' && curr_state == 'm') msglen += -log(prDM);
            else if (prev_state == 'd' && curr_state == 'd') msglen += -log(prDD);
            else if (prev_state == 'd' && curr_state == 'i') msglen += -log(prDI);
            else if (prev_state == 'i' && curr_state == 'm') msglen += -log(prIM);
            else if (prev_state == 'i' && curr_state == 'd') msglen += -log(prID);
            else if (prev_state == 'i' && curr_state == 'i') msglen += -log(prII);
        }
        prev_state = curr_state;
    }
   
    #ifdef DEBUG
    cout << "I(A)       = " << nits2bits(part1) << endl
         << "I(<S,T>|A) = " << nits2bits(part2) << endl
         << "I(A,<S,T>) = " << nits2bits(msglen) << endl;
    cout << "----------------------------+=o0o=+-------------------------------" << endl;
    #endif

    return msglen;
}

double threeStateDPAClass_t::evalNegLogMatchesOnly_fullAlignment(string fsastr, vector<string> A, int n) {
    
    assert(fsastr.length() == A.size());
    assert(n >= MARKOV_TIME_LLIMIT);
    assert(n <= MARKOV_TIME_ULIMIT);

    double msglen = log(MARKOV_TIME_ULIMIT - MARKOV_TIME_LLIMIT + 1);
    msglen += wtCodeLen_nits(fsastr.length());
   
    for (int l = 0; l < A.size(); l++) {
        char curr_state = fsastr[l];  
        if (curr_state == 'm') {
            assert(A[l][0]!='-'&&A[l][1]!='-');
            char S_i = A[l][0];
            char T_j = A[l][1];
            double matchPr = submat.jointpr(S_i,T_j,n);
            msglen += -log(matchPr);
        }
        else if (curr_state == 'd' || curr_state == 'i') {
            continue;
        }
        else {
            assert(!"Non m/i/d state found in fsastr");
        }
       
    }
   
    #ifdef DEBUG
    cout << "I(A,<S,T>) = " << nits2bits(msglen) << endl;
    cout << "----------------------------+=o0o=+-------------------------------" << endl;
    #endif

    return msglen;
}

void threeStateDPAClass_t::estimate3StateMachineParams_dirichlet( int N) {
    double prMM = submat.getprMM(N);
    double prMI = 0.5*(1.0-prMM);
    double prMD = 0.5*(1.0-prMM);
    fsaPr["MM"] = prMM;
    fsaPr["MI"] = prMI;
    fsaPr["MD"] = prMD;

    //double prII = Imode[0];
    double prII = submat.getprII(N);
    double prDD = prII;
    //double prIM = Imode[1];
    double prIM = submat.getprIM(N);
    double prDM = prIM;
    double prID = 1.0-prII-prIM;
    double prDI = prID;

    fsaPr["IM"] = prIM;
    fsaPr["II"] = prII;
    fsaPr["ID"] = prID;

    fsaPr["DM"] = prDM;
    fsaPr["DI"] = prDI;
    fsaPr["DD"] = prDD;

    assert(fsaPr["MM"]+fsaPr["MI"]+fsaPr["MD"] > 0.999999999999999);
    assert(fsaPr["IM"]+fsaPr["II"]+fsaPr["ID"] > 0.999999999999999);
    assert(fsaPr["DM"]+fsaPr["DI"]+fsaPr["DD"] > 0.999999999999999);

    /* //Read comment below
    paramStatementMsglen += 0; //mode params already in the codebook
    */

}

double threeStateDPAClass_t::getOptAlignMsgLen() {
  return algnModelMsgLen;
}

string threeStateDPAClass_t::getOptAlignFSAstr() {
  return optAlgnFSAStr;
}

double threeStateDPAClass_t::fsa2IA(string fsastr) { 
    double fsastrStringMsgLen = 0;
    int order = 1;
    int alphabetSize = 3;
    for (size_t i = 0; i < order; i++) {
        fsastrStringMsgLen += -log((double)1.0/alphabetSize); 
    }
    for (size_t i = order; i < fsastr.length(); i++) {
        if (fsastr[i-1] == 'm' && fsastr[i] == 'm') {
        fsastrStringMsgLen += -log(fsaPr["MM"]);
        }
        else if (fsastr[i-1] == 'm' && fsastr[i] == 'd') {
        fsastrStringMsgLen += -log(fsaPr["MD"]);
        }
        else if (fsastr[i-1] == 'm' && fsastr[i] == 'i') {
        fsastrStringMsgLen += -log(fsaPr["MI"]);
        }
        else if (fsastr[i-1] == 'd' && fsastr[i] == 'm') {
        fsastrStringMsgLen += -log(fsaPr["DM"]);
        }
        else if (fsastr[i-1] == 'd' && fsastr[i] == 'd') {
        fsastrStringMsgLen += -log(fsaPr["DD"]);
        }
        else if (fsastr[i-1] == 'd' && fsastr[i] == 'i') {
        fsastrStringMsgLen += -log(fsaPr["DI"]);
        }
        else if (fsastr[i-1] == 'i' && fsastr[i] == 'm') {
        fsastrStringMsgLen += -log(fsaPr["IM"]);
        }
        else if (fsastr[i-1] == 'i' && fsastr[i] == 'd') {
        fsastrStringMsgLen += -log(fsaPr["ID"]);
        }
        else if (fsastr[i-1] == 'i' && fsastr[i] == 'i') {
        fsastrStringMsgLen += -log(fsaPr["II"]);
        }
    }
    double IA =  wtCodeLen_nits(fsastr.length()) + paramStatementMsglen + fsastrStringMsgLen;
    //cout << "I(A)         = " << setw(9) << setprecision(3) << fixed << nits2bits(IA) << " bits" << endl;
    return IA;
}

double threeStateDPAClass_t::fsa2ISTgA(string fsastr) { 
    estimateParams(fsastr);
    double msglen = 0;
    int order = nullMdl.getOrder();
    int alphabetSize = nullMdl.getAlphabetSize();
    int iS = 0, jT = 0;
    for (size_t k = 0; k < fsastr.length(); k++) {
        if (fsastr[k] == 'd') {
            iS++;
        }
        else if (fsastr[k] == 'i') {
            jT++;
        }
        else if (fsastr[k] == 'm') {
            char S_i = S[iS];
            char T_j = T[jT];
            double matchPr = submat.jointpr(S_i,T_j,inferredSubmat_N);
            msglen += -log(matchPr);
            iS++;
            jT++;
        }
        else {
            cout << fsastr[k] << " " << "Something wierd!" << endl;
            assert(1);
        }
    }
    double ISTgA =  msglen;
    //cout << "I(<S,T>|A)   = " << setw(9) << setprecision(3) << fixed << nits2bits(ISTgA) << " bits" << endl;
    return ISTgA;
}

double threeStateDPAClass_t::fsa2ISTgAMatchOnly(string fsastr) { 
    estimateParams(fsastr);

    double msglen = 0;
    int order = nullMdl.getOrder();
    int alphabetSize = nullMdl.getAlphabetSize();
    int iS = 0, jT = 0;
    for (size_t k = 0; k < fsastr.length(); k++) {
        if (fsastr[k] == 'd') {
            if (k < order) msglen += -log((double)1/alphabetSize);
            else msglen += -log(nullMdl.pr(S.substr(iS-order,order+1)));
            iS++;
        }
        else if (fsastr[k] == 'i') {
            if (k < order) msglen += -log((double)1/alphabetSize);
            else msglen += -log(nullMdl.pr(T.substr(jT-order,order+1)));
            jT++;
        }
        else if (fsastr[k] == 'm') {
            char S_i = S[iS];
            char T_j = T[jT];
            double matchPr = submat.jointpr(S_i,T_j,inferredSubmat_N);
            msglen += -log(matchPr);
            iS++;
            jT++;
        }
        else {
            cout << fsastr[k] << " " << "Something wierd!" << endl;
            assert(1);
        }
    }
    double ISTgA =  msglen;
    //cout << "I(<S,T>|A)   = " << setw(9) << setprecision(3) << fixed << nits2bits(ISTgA) << " bits" << endl;
    return ISTgA;
}

int threeStateDPAClass_t::getInferredSubmatN() {
    return inferredSubmat_N;
}
